/*******************************************************************************
 * Copyright 2015, 2016 Francesco Benincasa.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *******************************************************************************/
package com.abubusoft.kripton.binder.json;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.abubusoft.kripton.BinderJsonReader;
import com.abubusoft.kripton.BinderOptions;
import com.abubusoft.kripton.binder.json.internal.JSONArray;
import com.abubusoft.kripton.binder.json.internal.JSONObject;
import com.abubusoft.kripton.binder.schema.ElementSchema;
import com.abubusoft.kripton.binder.schema.ElementSchema.MapInfo;
import com.abubusoft.kripton.binder.schema.MappingSchema;
import com.abubusoft.kripton.binder.transform.Transformer;
import com.abubusoft.kripton.common.Pair;
import com.abubusoft.kripton.common.StringUtil;
import com.abubusoft.kripton.common.TypeReflector;
import com.abubusoft.kripton.exception.MappingException;
import com.abubusoft.kripton.exception.ReaderException;

/**
 * Json reader
 * 
 * @author xcesco
 * 
 */
public class JsonReader implements BinderJsonReader {

	/**
	 * format of json reader
	 */
	private BinderOptions format;

	public JsonReader() {
		this(new BinderOptions());
	}

	public JsonReader(BinderOptions format) {
		this.format = format;
	}

	public <T> T read(Class<? extends T> type, InputStream source) throws ReaderException, MappingException {

		try {
			return this.read(type, new InputStreamReader(source, format.getEncoding()));
		} catch (UnsupportedEncodingException e) {
			throw new ReaderException("Encoding is not supported", e);
		}

	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public <T> T read(Class<? extends T> type, String source) throws ReaderException, MappingException {
		// if empty input, exit immediately
		if (StringUtil.isEmpty(source)) {
			return null;
		}

		validate(type, source);

		try {
			Constructor con = TypeReflector.getConstructor(type);
			Object instance = con.newInstance();

			JSONObject jsonObj = new JSONObject(source);

			this.readObject(instance, jsonObj);

			return (T) instance;
		} catch (ReaderException re) {
			throw re;
		} catch (MappingException me) {
			throw me;
		} catch (Exception e) {
			throw new ReaderException("Error to read object " + e.getMessage(), e);
		}
	}

	public <T> T read(Class<? extends T> type, Reader source) throws ReaderException, MappingException {
		if (source == null) {
			throw new ReaderException("Cannot read, reader is null!");
		}

		try {
			return this.read(type, StringUtil.reader2String(source));
		} catch (IOException e) {
			throw new ReaderException("IO error!", e);
		}

	}

	// private String reader2String(Reader source) throws IOException {
	// char[] cbuf = new char[65535];
	// StringBuffer stringbuf = new StringBuffer();
	//
	// int count = 0;
	// while ((count = source.read(cbuf, 0, 65535)) != -1) {
	// stringbuf.append(cbuf, 0, count);
	// }
	//
	// return stringbuf.toString();
	// }

	private <T> void validate(Class<? extends T> type, String source) throws ReaderException {
		if (type == null) {
			throw new ReaderException("Cannot read, type is null!");
		}

		if (Transformer.isPrimitive(type)) {
			throw new ReaderException("Can not read primitive or enum type object, " + "only Nano bindable complex type object can be accepted!");
		}
	}

	private void readObject(Object instance, JSONObject jsonObj) throws Exception {
		MappingSchema ms = MappingSchema.fromObject(instance);

		this.readElement(instance, jsonObj, ms);
	}

	private void readElement(Object instance, JSONObject jsonObj, MappingSchema ms) throws Exception {
		Map<String, ElementSchema> xml2SchemaMapping = ms.getXml2SchemaMapping();
		readElementInternal(instance, jsonObj, xml2SchemaMapping);
		Map<String, ElementSchema> xmlWrapper2SchemaMapping = ms.getXmlWrapper2SchemaMapping();
		readElementInternal(instance, jsonObj, xmlWrapper2SchemaMapping);
	}

	private void readElementInternal(Object instance, JSONObject jsonObj, Map<String, ElementSchema> map) throws Exception {
		Class<?> type;
		for (String name : map.keySet()) {
			Object jsonValue = jsonObj.opt(name);
			if (jsonValue != null) {
				ElementSchema es = map.get(name);

				if (!es.getJsonInfo().enabled)
					continue;

				Field field = es.getField();
				type = es.getFieldType();

				switch (es.getType()) {
				case LIST:
					// List
					if (jsonValue instanceof JSONArray) {
						JSONArray jsonArray = (JSONArray) jsonValue;
						if (jsonArray.length() > 0) {
							List<?> list = (List<?>) field.get(instance);
							if (list == null) {
								list = new ArrayList<>();
								field.set(instance, list);
							}

							readList(instance, type, list, jsonArray);
						}
					}
					break;
				case SET:
					// Set
					if (jsonValue instanceof JSONArray) {
						JSONArray jsonArray = (JSONArray) jsonValue;
						if (jsonArray.length() > 0) {
							readSet(instance, type, field, jsonArray);
						}
					}

					break;
				case ARRAY:
					// Array
					if (jsonValue instanceof JSONArray) {
						JSONArray jsonArray = (JSONArray) jsonValue;
						if (jsonArray.length() > 0) {
							readArray(instance, type, field, jsonArray);
						}
					}
					break;
				case MAP:
					// Array
					if (jsonValue instanceof JSONArray) {
						JSONArray jsonArray = (JSONArray) jsonValue;
						if (jsonArray.length() > 0) {
							readMap(instance, type, field, jsonArray, es.getMapInfo());
						}
					}
					break;
				case ELEMENT:
					if (jsonValue instanceof JSONArray) {
						jsonValue = ((JSONArray) jsonValue).get(0);
					}
					if (Transformer.isPrimitive(type)) {
						if (!(jsonValue instanceof JSONObject) && !(jsonValue instanceof JSONArray) && !(jsonValue instanceof JSONObject.Null)) {
							Object value = Transformer.read(String.valueOf(jsonValue), type);
							field.set(instance, value);
						}
					} else if (jsonValue instanceof JSONObject) {
						Constructor<?> con = TypeReflector.getConstructor(type);
						Object subObj = con.newInstance();
						field.set(instance, subObj);
						this.readObject(subObj, (JSONObject) jsonValue);
					}
					break;
				}
			}
		}
	}

	private void readMap(Object instance, Class<?> type, Field field, JSONArray jsonArray, MapInfo mapInfo) throws Exception {
		Map<?, ?> map = (Map<?, ?>) field.get(instance);
		if (map == null) {
			map = new LinkedHashMap<>();
			field.set(instance, map);
		}
		Object entry;

		String keyName = mapInfo.keyName;
		Class<?> keyClazz = mapInfo.keyClazz;
		String valueName = mapInfo.valueName;
		Class<?> valueClazz = mapInfo.valueClazz;

		for (int i = 0; i < jsonArray.length(); i++) {
			entry = jsonArray.get(i);
			if (entry instanceof JSONObject) {
				readMapEntry(map, entry, keyName, keyClazz, valueName, valueClazz);
			} else {
				throw new MappingException("Badformat for map field " + field.getName() + " of class " + instance.getClass().getName());
			}
		}

	}

	/**
	 * @param map
	 * @param entry
	 * @param keyName
	 * @param keyClazz
	 * @param valueName
	 * @param valueClazz
	 * @throws Exception
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	@SuppressWarnings("unchecked")
	private void readMapEntry(@SuppressWarnings("rawtypes") Map map, Object entry, String keyName, Class<?> keyClazz, String valueName, Class<?> valueClazz) throws Exception, InstantiationException, IllegalAccessException {
		Object objectValue;
		Object objectKey;
		Object value;
		Object key;
		JSONObject jsonEntry;
		MappingSchema msKey;
		MappingSchema msValue;
		jsonEntry = (JSONObject) entry;
		key = jsonEntry.get(keyName);

		if (Transformer.isPrimitive(key.getClass())) {
			// simple type
			objectKey = Transformer.read(String.valueOf(key), keyClazz);
		} else {
			objectKey = keyClazz.newInstance();
			msKey = MappingSchema.fromClass(keyClazz);
			readElementInternal(objectKey, (JSONObject) key, msKey.getField2SchemaMapping());
		}

		value = jsonEntry.opt(valueName);
		if (value == null) {
			objectValue = null;
		} else if (Transformer.isPrimitive(value.getClass())) {
			// simple type
			objectValue = Transformer.read(String.valueOf(value), valueClazz);
		} else {
			objectValue = valueClazz.newInstance();
			msValue = MappingSchema.fromClass(valueClazz);
			readElementInternal(objectValue, (JSONObject) value, msValue.getField2SchemaMapping());
		}

		map.put(objectKey, objectValue);
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void readSet(Object instance, Class<?> type, Field field, JSONArray jsonArray) throws Exception {
		if (Transformer.isPrimitive(type)) {
			for (int i = 0; i < jsonArray.length(); i++) {
				Object jsonValue = jsonArray.get(i);
				if (!(jsonValue instanceof JSONObject) && !(jsonValue instanceof JSONArray)) {
					Object value = Transformer.read(String.valueOf(jsonValue), type);
					Set set = (Set) field.get(instance);
					if (set == null) {
						set = new LinkedHashSet();
						field.set(instance, set);
					}
					set.add(value);
				}
			}
		} else { // Object
			for (int i = 0; i < jsonArray.length(); i++) {
				Object jsonValue = jsonArray.get(i);
				if (jsonValue instanceof JSONObject) {
					Constructor con = TypeReflector.getConstructor(type);
					Object subObj = con.newInstance();
					Set set = (Set) field.get(instance);
					if (set == null) {
						set = new LinkedHashSet();
						field.set(instance, set);
					}
					set.add(subObj);
					this.readObject(subObj, (JSONObject) jsonValue);
				}
			}
		}

	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void readList(Object instance, Class<?> type, Collection list, JSONArray jsonArray) throws Exception {
		if (Transformer.isPrimitive(type)) {
			for (int i = 0; i < jsonArray.length(); i++) {
				Object jsonValue = jsonArray.get(i);
				if (!(jsonValue instanceof JSONObject) && !(jsonValue instanceof JSONArray)) {
					Object value = Transformer.read(String.valueOf(jsonValue), type);
					list.add(value);
				}
			}
		} else { // Object
			for (int i = 0; i < jsonArray.length(); i++) {
				Object jsonValue = jsonArray.get(i);
				if (jsonValue instanceof JSONObject) {
					Constructor con = TypeReflector.getConstructor(type);
					Object subObj = con.newInstance();

					list.add(subObj);
					this.readObject(subObj, (JSONObject) jsonValue);
				}
			}
		}
	}

	@SuppressWarnings({ "rawtypes" })
	private void readArray(Object instance, Class<?> type, Field field, JSONArray jsonArray) throws Exception {
		Object array = Array.newInstance(type, jsonArray.length());
		field.set(instance, array);

		Object subObj;
		Object jsonValue;
		for (int i = 0; i < jsonArray.length(); i++) {
			jsonValue = jsonArray.get(i);
			if (jsonValue instanceof JSONObject) {
				Constructor con = TypeReflector.getConstructor(type);
				subObj = con.newInstance();
				Array.set(array, i, subObj);
				this.readObject(subObj, (JSONObject) jsonValue);
			} else if (jsonValue instanceof JSONArray) {
				// TODO to implement
			} else {
				subObj = Transformer.read(String.valueOf(jsonValue), type);
				Array.set(array, i, subObj);

			}
		}
	}

	@Override
	public <E> List<E> readList(Class<E> type, String input) throws ReaderException {
		JSONArray array = new JSONArray(input);
		List<E> result = new ArrayList<E>();

		try {
			readList(input, type, result, array);

			return result;
		} catch (Exception e) {
			e.printStackTrace();
			throw (new ReaderException(e.getMessage()));
		}
	}

	@Override
	public <E> List<E> readList(Class<E> type, InputStream source) throws ReaderException {
		try {
			return this.readList(type, new InputStreamReader(source, format.getEncoding()));
		} catch (UnsupportedEncodingException e) {
			throw new ReaderException("Encoding is not supported", e);
		}
	}

	@Override
	public <E> List<E> readList(Class<E> type, Reader source) throws ReaderException, MappingException {
		if (source == null) {
			throw new ReaderException("Cannot read, reader is null!");
		}

		try {
			return this.readList(type, StringUtil.reader2String(source));
		} catch (IOException e) {
			throw new ReaderException("IO error!", e);
		}

	}

	@Override
	public <L extends Collection<E>, E> L readCollection(L list, Class<E> type, String input) throws ReaderException {
		JSONArray array = new JSONArray(input);
		L result = list;

		try {
			readList(input, type, result, array);

			return result;
		} catch (Exception e) {
			e.printStackTrace();
			throw (new ReaderException(e.getMessage()));
		}
	}

	@Override
	public <L extends Collection<E>, E> L readCollection(L collection, Class<E> type, InputStream source) throws ReaderException {
		try {
			return this.readCollection(collection, type, new InputStreamReader(source, format.getEncoding()));
		} catch (UnsupportedEncodingException e) {
			throw new ReaderException("Encoding is not supported", e);
		}
	}

	@Override
	public <L extends Collection<E>, E> L readCollection(L collection, Class<E> type, Reader source) throws ReaderException, MappingException {
		if (source == null) {
			throw new ReaderException("Cannot read, reader is null!");
		}

		try {
			return this.readCollection(collection, type, StringUtil.reader2String(source));
		} catch (IOException e) {
			throw new ReaderException("IO error!", e);
		}

	}

	protected boolean isList(@SuppressWarnings("rawtypes") Class clazz) {
		if (clazz == null)
			return false;

		if (List.class.isAssignableFrom(clazz)) {
			return true;
		}

		return false;
	}

	@SuppressWarnings("rawtypes")
	protected Pair<Class, Class> getListElementClass(Class clazz) {
		if (List.class.isAssignableFrom(clazz)) {
			final ParameterizedType genericSuperclass = (ParameterizedType) clazz.getGenericSuperclass();

			Class<?> elementType = (Class<?>) genericSuperclass.getActualTypeArguments()[0];

			return new Pair<Class, Class>(clazz, elementType);
		}

		return null;
	}

	@Override
	public <M extends Map<K, V>, K, V> M readMap(M map, Class<K> keyType, Class<V> valueType, String source) throws ReaderException {
		JSONArray jsonArray = new JSONArray(source);
		String keyName = "key";
		String valueName = "value";
		Object entry;

		try {
			for (int i = 0; i < jsonArray.length(); i++) {
				entry = jsonArray.get(i);
				if (entry instanceof JSONObject) {

					readMapEntry(map, entry, keyName, keyType, valueName, valueType);

				} else {
					throw new MappingException("Badformat for map");
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			throw new ReaderException("Badformat for map");
		}
		
		return map;
	}

	@Override
	public <M extends Map<K, V>, K, V> M readMap(M map, Class<K> keyType, Class<V> valueType, InputStream source) throws ReaderException {
		try {
			return readMap(map, keyType, valueType, new InputStreamReader(source, format.getEncoding()));
		} catch (UnsupportedEncodingException e) {
			throw new ReaderException("Encoding is not supported", e);
		}
	}

	@Override
	public <M extends Map<K, V>, K, V> M readMap(M map, Class<K> keyType, Class<V> valueType, Reader source) throws ReaderException, MappingException {
		if (source == null) {
			throw new ReaderException("Cannot read, reader is null!");
		}

		try {
			return this.readMap(map, keyType, valueType, StringUtil.reader2String(source));
		} catch (IOException e) {
			throw new ReaderException("IO error!", e);
		}
	}

}
