/*******************************************************************************
 * Copyright 2015, 2016 Francesco Benincasa.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *******************************************************************************/
package com.abubusoft.kripton.binder.xml;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.abubusoft.kripton.exception.ReaderException;
import com.abubusoft.kripton.binder.schema.ElementSchema;
import com.abubusoft.kripton.binder.schema.MappingSchema;
import com.abubusoft.kripton.binder.schema.TypeElementSchema;
import com.abubusoft.kripton.binder.schema.SchemaArray;
import com.abubusoft.kripton.binder.transform.Transformer;
import com.abubusoft.kripton.binder.xml.internal.MapEntry;
import com.abubusoft.kripton.binder.xml.internal.MapEntryImpl;
import com.abubusoft.kripton.common.StringUtil;
import com.abubusoft.kripton.common.TypeReflector;

/**
 * SAX handler implementation for XmlSaxReader
 * 
 * @author bulldog
 * @author xcesco
 * 
 */
class XmlReaderHandler extends DefaultHandler {

	private XmlReaderHelper helper;

	@Override
	public void endDocument() throws SAXException {
	}

	public XmlReaderHandler(XmlReaderHelper helper) {
		this.helper = helper;
	}

	private void populateAttributes(Object obj, Attributes attrs, MappingSchema ms) throws Exception {
		Map<String, ElementSchema> attributeSchemaMapping = ms.getXml2AttributeSchemaMapping();

		for (int index = 0; index < attrs.getLength(); index++) {
			String attrName = attrs.getLocalName(index);

			ElementSchema as = attributeSchemaMapping.get(attrName);
			if (as == null)
				continue;

			if (!as.getXmlInfo().enabled)
				continue;

			String attrValue = attrs.getValue(index);

			Object value = Transformer.read(attrValue, as.getFieldType());

			Field field = as.getField();
			field.set(obj, value);
		}
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public void startElement(String uri, String localName, String name, Attributes attrs) throws SAXException {
		try {
			Object obj = helper.valueStack.peek();
			MappingSchema ms = MappingSchema.fromObject(obj);

			Map<String, ElementSchema> xmlWrapper2SchemaMapping = ms.getXmlWrapper2SchemaMapping();
			if (xmlWrapper2SchemaMapping.containsKey(localName)) {
				return;
			}

			helper.depth++;
			// clear the textBuilder
			helper.clearTextBuffer();

			if (helper.depth > helper.valueStack.size() + 1) {
				// unexpected xml element, just ignore
				return;
			}

			if (helper.isRoot()) { // first time root element mapping
				TypeElementSchema res = ms.getRootElementSchema();
				String xmlName = res.xmlInfo.getName();

				// simple validation only for root element
				if (!xmlName.equalsIgnoreCase(localName)) {
					throw new ReaderException("Root element name mismatch, " + localName + " != " + xmlName);
				}

				if (attrs != null && attrs.getLength() > 0) {
					this.populateAttributes(obj, attrs, ms);
				}
			} else {
				// sub element mapping
				Class<?> type = null;
				if (ms.xmlInfo.isMapEntryStub()) {
					MapEntry mapStrategy = (MapEntry) obj;

					if (mapStrategy.isKey(localName)) {
						type = mapStrategy.getMapInfo().keyClazz;
					} else if (mapStrategy.isValue(localName)) {
						type = mapStrategy.getMapInfo().valueClazz;
					}

					if (!Transformer.isPrimitive(type)) {
						Object newObj = type.newInstance();
						helper.valueStack.push(newObj);
					}
				} else {
					Map<String, ElementSchema> xml2SchemaMapping = ms.getXml2SchemaMapping();
					ElementSchema schema = xml2SchemaMapping.get(localName);

					// detect type
					type = schema.getFieldType();

					if (!Transformer.isPrimitive(type)) {
						Object newObj = null;
						if (schema.isMap()) {
							Map map = (Map) schema.getField().get(obj);
							if (map == null) {
								map = new LinkedHashMap();
								schema.getField().set(obj, map);
							}

							switch (schema.getMapInfo().entryStrategy) {
							case ATTRIBUTES: {
								MapEntryImpl mapPolicy = new MapEntryImpl();
								mapPolicy.set(map, schema.getMapInfo());

								if (attrs != null && attrs.getLength() > 0) {
									this.populateAttributesForMap(mapPolicy, attrs);
								}

								if (mapPolicy.isEntryReady()) {
									map.put(mapPolicy.getEntryKey(), mapPolicy.getEntryValue());
									mapPolicy.clear();
								}
							}
								// exit, DO NOT PUSH OBJECT!
								return;
							case ELEMENTS: {
								MapEntryImpl mapPolicy = new MapEntryImpl();
								mapPolicy.set(map, schema.getMapInfo());

								newObj = mapPolicy;
							}
								break;
							}

						} else {
							MappingSchema newMs = MappingSchema.fromClass(type);
							Constructor con = null;
							try {
								con = TypeReflector.getConstructor(type);
							} catch (NoSuchMethodException nsme) {
								throw new ReaderException("No-arg constructor is missing, type = " + type.getName());
							}
							newObj = con.newInstance();
							if (attrs != null && attrs.getLength() > 0) {
								this.populateAttributes(newObj, attrs, newMs);
							}
						}

						helper.valueStack.push(newObj);
					}

				}
			}

		} catch (Exception ex) {
			throw new SAXException("Reading exception in startElement, " + ex.getMessage(), ex);
		}
	}

	private void populateAttributesForMap(MapEntryImpl mapPolicy, Attributes attrs) throws Exception {
		for (int index = 0; index < attrs.getLength(); index++) {
			String attrName = attrs.getLocalName(index);
			String attrValue = attrs.getValue(index);

			if (mapPolicy.isValue(attrName)) {
				Object value = Transformer.read(attrValue, mapPolicy.getMapInfo().valueClazz);
				mapPolicy.setEntryValue(value);
			} else if (mapPolicy.isKey(attrName)) {
				Object value = Transformer.read(attrValue, mapPolicy.getMapInfo().keyClazz);
				mapPolicy.setEntryKey(value);
			}
		}
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public void endElement(String uri, String localName, String name) throws SAXException {
		try {
			{
				Object obj = helper.valueStack.peek();
				MappingSchema ms = MappingSchema.fromObject(obj);

				Map<String, ElementSchema> xmlWrapper2SchemaMapping = ms.getXmlWrapper2SchemaMapping();
				if (xmlWrapper2SchemaMapping.containsKey(localName)) {
					return;
				}

				SchemaArray lastArray = helper.arrayStack.size() > 0 ? helper.arrayStack.peek() : null;
				if (lastArray != null && lastArray.value0.getField().getDeclaringClass() == obj.getClass() && !lastArray.value0.getName().equals(localName)) {
					ElementSchema es = lastArray.value0;

					Field field = es.getField();

					int n = lastArray.value1.size();
					Object value = Array.newInstance(es.getFieldType(), lastArray.value1.size());
					for (int i = 0; i < n; i++) {
						Array.set(value, i, lastArray.value1.get(i));
					}

					if (!field.isAccessible()) {
						field.setAccessible(true);
					}

					if (es.getXmlInfo().enabled) {
						field.set(obj, value);
					}
					helper.arrayStack.pop();
				}

			}

			if (helper.depth > helper.valueStack.size() + 1) {
				// unexpected xml element, just ignore
				helper.depth--;
				return;
			} else if (helper.depth == helper.valueStack.size() + 1) {
				// handle primitive field
				// Object obj = helper.valueStack.peek();
				Object obj = helper.valueStack.peek();
				MappingSchema ms = MappingSchema.fromObject(obj);

				Map<String, ElementSchema> xml2SchemaMapping = ms.getXml2SchemaMapping();
				ElementSchema schema = xml2SchemaMapping.get(localName);

				if (ms.xmlInfo.isMapEntryStub()) {
					MapEntry mapStrategy = (MapEntry) obj;

					String xmlData = helper.textBuilder.toString();
					if (!StringUtil.isEmpty(xmlData)) {
						Class<?> type;
						Object value;

						if (mapStrategy.isKey(localName)) {
							type = mapStrategy.getMapInfo().keyClazz;
							if (Transformer.isPrimitive(type)) {
								value = Transformer.read(xmlData, type);
								mapStrategy.setEntryKey(value);
							}

						} else if (mapStrategy.isValue(localName)) {
							type = mapStrategy.getMapInfo().valueClazz;
							if (Transformer.isPrimitive(type)) {
								value = Transformer.read(xmlData, type);
								mapStrategy.setEntryValue(value);
							}
						}
					}
					// skip depht decrement
					helper.depth--;
					return;
				} else {

					Field field = schema.getField();
					String xmlData = helper.textBuilder.toString();
					if (!StringUtil.isEmpty(xmlData)) {
						if (ms.xmlInfo.isMapEntryStub()) {
							MapEntry mapStrategy = (MapEntry) obj;
							Class<?> type = null;
							Object value = null;

							if (mapStrategy.isKey(localName)) {
								type = mapStrategy.getMapInfo().keyClazz;
								value = Transformer.read(xmlData, type);
								mapStrategy.setEntryKey(value);
							} else if (mapStrategy.isValue(localName)) {
								type = mapStrategy.getMapInfo().valueClazz;
								value = Transformer.read(xmlData, type);
								mapStrategy.setEntryValue(value);
							}

							if (mapStrategy.isEntryReady()) {
								mapStrategy.getMap().put(mapStrategy.getEntryKey(), mapStrategy.getEntryValue());
								mapStrategy.clear();
							}

						} else {

							if (schema.isList()) {
								Class<?> paramizedType = schema.getFieldType();
								Object value = Transformer.read(xmlData, paramizedType);
								List list = (List) field.get(obj);
								if (list == null) {
									list = new ArrayList();

									if (schema.getXmlInfo().enabled) {
										field.set(obj, list);
									}
								}
								list.add(value);
							} else if (schema.isSet()) {
								Class<?> paramizedType = schema.getFieldType();
								Object value = Transformer.read(xmlData, paramizedType);
								Set set = (Set) field.get(obj);
								if (set == null) {
									set = new LinkedHashSet();
									if (schema.getXmlInfo().enabled) {
										field.set(obj, set);
									}
								}
								set.add(value);
							} else if (schema.isArray() && schema.getFieldType() != Byte.TYPE) {
								SchemaArray schemaArray = helper.arrayStack.size() > 0 ? helper.arrayStack.peek() : null;

								if (schemaArray == null || schemaArray.value0 != schema) {
									schemaArray = new SchemaArray(schema, new ArrayList());
									helper.arrayStack.add(schemaArray);
								}
								Class<?> paramizedType = schema.getFieldType();
								Object value;
								// --------
								if (Transformer.isPrimitive(paramizedType)) {
									value = Transformer.read(xmlData, paramizedType);
								} else {
									value = obj;
								}
								// --------

								schemaArray.value1.add(value);

							} else {
								Object value = Transformer.read(xmlData, field.getType());
								if (schema.getXmlInfo().enabled) {
									field.set(obj, value);
								}
							}
						}
					}
				}
			} else if (helper.depth == helper.valueStack.size()) {
				// handle object field
				Object obj = helper.valueStack.pop();
				MappingSchema ms = MappingSchema.fromObject(obj);

				ElementSchema vs = ms.getValueSchema();

				if (vs != null) {
					if (vs.getXmlInfo().enabled) {
						helper.depth--;
						return;
					}

					Field field = vs.getField();
					String xmlData = helper.textBuilder.toString();
					if (!StringUtil.isEmpty(xmlData)) {
						Object value = Transformer.read(xmlData, vs.getFieldType());
						if (vs.getXmlInfo().enabled) {
							field.set(obj, value);
						}
					}
				}

				if (helper.valueStack.size() == 0) { // the end
					helper.valueStack.push(obj);
					helper.depth--;
					return;
				}

				// retrieve parent object
				Object parentObj = helper.valueStack.peek();
				MappingSchema parentMs = MappingSchema.fromObject(parentObj);
				Map<String, ElementSchema> parentXml2SchemaMapping = parentMs.getXml2SchemaMapping();

				ElementSchema schema = parentXml2SchemaMapping.get(localName);
				if (schema != null && schema.getXmlInfo().enabled ) {
					ElementSchema es = (ElementSchema) schema;
					Field field = es.getField();
					if (ms.xmlInfo.isMapEntryStub()) {
						// do nothing... all is done!
						MapEntry mapObj = (MapEntry) obj;
						if (mapObj.isKeyPresent()) {
							mapObj.getMap().put(mapObj.getEntryKey(), mapObj.getEntryValue());
						}
					} else if (parentMs.xmlInfo.isMapEntryStub()) {
						MapEntry mapStrategy = (MapEntry) parentObj;
						Object value = null;

						if (mapStrategy.isKey(localName)) {
							value = obj;
							mapStrategy.setEntryKey(value);
						} else if (mapStrategy.isValue(localName)) {
							value = obj;
							mapStrategy.setEntryValue(value);
						}

						if (mapStrategy.isEntryReady()) {
							mapStrategy.getMap().put(mapStrategy.getEntryKey(), mapStrategy.getEntryValue());
							mapStrategy.clear();
						}
					} else if (es.isList()) {
						List list = (List) field.get(parentObj);
						if (list == null) {
							list = new ArrayList();
							if (!field.isAccessible()) {
								field.setAccessible(true);
							}
							field.set(parentObj, list);
						}
						list.add(obj);
					} else if (es.isSet()) {
						Set set = (Set) field.get(parentObj);
						if (set == null) {
							set = new LinkedHashSet();
							if (!field.isAccessible()) {
								field.setAccessible(true);
							}
							field.set(parentObj, set);
						}
						set.add(obj);
					} else if (es.isArray() && es.getFieldType() != Byte.TYPE) {
						SchemaArray schemaArray = helper.arrayStack.size() > 0 ? helper.arrayStack.peek() : null;

						if (schemaArray == null || schemaArray.value0 != es) {
							schemaArray = new SchemaArray(es, new ArrayList());
							helper.arrayStack.add(schemaArray);
						}
						Class<?> paramizedType = es.getFieldType();
						// Object value = Transformer.read(xmlData,
						// paramizedType);
						Object value = null;
						// --------
						if (Transformer.isPrimitive(paramizedType)) {
							// value = Transformer.read(xmlData, paramizedType);
							// not here
						} else {
							value = obj;
						}
						// --------

						schemaArray.value1.add(value);
					} else {
						field.set(parentObj, obj);
					}
				}

			}

			helper.depth--;
		} catch (Exception ex) {
			throw new SAXException("Reading Exception in endElement, " + ex.getMessage(), ex);
		}
	}

	public void characters(char[] ch, int start, int length) throws SAXException {
		String text = new String(ch, start, length);
		helper.textBuilder.append(text);
	}

}
